'------------------------------------------------------------------
' si5351a DDS Oscillator with I2C LCD display
' Author: Andrew Woodfield ZL2PD
' Date: March 2016
' File Name: si85vfoxx.bas     (where xx is version number)
'
' Released under the terms and conditions of the
' Creative Commons Public License ("CCPL")
' See creativecommons.org for details
'
' Commercial use is prohibited without written permission
' from the author
'
'------------------------------------------------------------------
' Adapted from boardtest11.bas:
'
' V01: initial code
'
'------------------------------------------------------------------
' Credits:
'
' I2C LCD code adapted from www.ne.jp/asahi/shared/o-family/ElecRoom/ElecMAIN.htm
' si5351 code is adapted from OE1CGS and Jason Milldrum C code
'

'
'------------------------------------------------------------------
' Strpped down for basic operation and control over RS232
' by Burkhard Kainka, DK7JD, March 2019

'------------------------------------------------------------------

'------------------------------------------------------------------
' Compiler Directives (some fundamental hardware issues)

$regfile = "attiny85.dat"
$crystal = 8000000
$hwstack = 128
$swstack = 64
$framesize = 128
'-------------------------------------------------------------------------------
' Declare Variables

Dim Temp As Byte                                            'temporary variable used in sendfreq and sbarplot routines


Dim Frequency As Dword                                      'four byte unsigned variable
Dim Gfreq As Dword                                          'calculates si5351 VFO output frequency
Dim Clk As Byte                                             'output 0/1/2

Dim Fvco As Dword                                           'si5351 vco frequency
Dim Outdivider As Dword                                     '4,6,8-900
Dim Rdivider As Byte                                        'Normally 1 unless freqeuncy<1MHz then 1,2,4,8,16,32,64 or 128 only
Dim Rbyte As Byte                                           'used as temp store before final data load to si5351
Dim Afactor As Dword
Dim Bfactor As Dword
Dim Fvalue As Single                                        'used to determine accuracy of fractional calculation
Dim Finteger As Dword

Dim Ms0p1 As Dword                                          '18-bit MS0_P1 si5351 parameters for Output 0
Dim M0p11 As Byte At Ms0p1 Overlay                          'mirrored variables for dds calculation
Dim M0p12 As Byte At Ms0p1 + 1 Overlay
Dim M0p13 As Byte At Ms0p1 + 2 Overlay
Dim M0p14 As Byte At Ms0p1 + 3 Overlay                      'most significant byte for ms0p1

'Note: MS0_P2 and MS0_P3 are hard-coded as 0 and 1 respectively so ignored during dimensioning
                                                              'si5351 parameters for PLLA
Dim Msnap1 As Dword                                         '18-bit variable for MSNA_P1
Dim Ms11 As Byte At Msnap1 Overlay                          'mirrored variables for dds calculation
Dim Ms12 As Byte At Msnap1 + 1 Overlay
Dim Ms13 As Byte At Msnap1 + 2 Overlay
Dim Ms14 As Byte At Msnap1 + 3 Overlay                      'most significant byte

Dim Msnap2 As Dword                                         '20-bit variable for MSNA_P2
Dim Ms21 As Byte At Msnap2 Overlay                          'mirrored variables for dds calculation
Dim Ms22 As Byte At Msnap2 + 1 Overlay
Dim Ms23 As Byte At Msnap2 + 2 Overlay
Dim Ms24 As Byte At Msnap2 + 3 Overlay                      'most significant byte

Dim Msnap3 As Dword                                         '20-bit variable for MSNA_P3
Dim Ms31 As Byte At Msnap3 Overlay                          'mirrored variables for dds calculation
Dim Ms32 As Byte At Msnap3 + 1 Overlay
Dim Ms33 As Byte At Msnap3 + 2 Overlay
Dim Ms34 As Byte At Msnap3 + 3 Overlay                      'most significant byte

Dim Regaddr As Byte , Regdata As Byte                       'for i2c calls
Dim Siaddr1 As Byte , Siaddr2 As Byte , Siaddr3 As Byte     'for writing to si5351 CLK0 and CLK1

Dim Zeile As String * 20
Dim Komm As String * 1
Dim L As Byte
Dim Freq As Word
Dim Freq2 As Word
Dim Ampl As Word
Dim Ampl2 As Word
Dim Ampl3 As Word
Dim Modus As Byte
Dim U As Word
Dim S As String * 10
Dim Dat As Word
Dim D As Word

' Declare Constants
Const Xtal = 25000000                                       '**NOTE: Value may be adjusted during alignment ***
Const Cfactor = 1048575
'Const Ifoffset = 8867000                                    'ifoffset=0 if no offset e.g. VFO used in SDR or DC receiver
'Const Bfofreq = 8870000                                     'bfo/cio frequencies but not currently supported in this version

'Const Midas_vdd = 3                                         'LCD display is powered from 3V rail
Open "comb.2:9600,8,n,1" For Input As #1

' Declare Subroutines
Declare Sub Si5351init                                      'initialises si5351
Declare Sub Setfreq                                         'output to CLK0/1/2 in Hz
Declare Sub Calculate                                       'converts current VFO or BFO freq to data for si5351
Declare Sub Sendfreq                                        'sends VFO or BFO/CIO freq data to si5351 CLK0 or CLK1 respectively
Declare Sub Siout(sireg As Byte , Sidata As Byte)           'for i2c write to si5351
                                      'plots current s-meter input voltage

'------------------------------------------------------------------------------------

Config Portb = Output                                       '

Config Sda = Portb.4
Config Scl = Portb.3
I2cinit
Waitms 50                                                   'for i2c to settle and for lcd RC reset time
Si5351init                                                  'initialises si5351
Waitms 500
Gfreq = 14296000
Clk = 1
Setfreq
Waitms 500

Do
  Dat = Ischarwaiting(#1)
  If Dat > 0 Then
   Input #1 , Zeile
   Komm = Left(zeile , 1)
   L = Len(zeile)
   L = L - 1
   Zeile = Right(zeile , L)

   If Komm = "x" Then                                       'Frequenz 0
     Freq = Val(zeile)
     Gfreq = Freq * 1000
     Clk = 0
     Setfreq
   End If

   If Komm = "y" Then                                       'Frequenz 1
     Freq = Val(zeile)
     Gfreq = Freq * 1000                                    '
     Clk = 1
     Setfreq
   End If
  End If
Loop


'**********************************************************************

Sub Setfreq
  Calculate                                                 'Gfreq to chip registers
  If Clk = 0 Then
    Siaddr1 = 26                                            'point to CLK0 base register addresses
    Siaddr2 = 42
    Siaddr3 = 44
    Sendfreq                                                'and send it to the si5351 CLK0
    Regaddr = 16
    Regdata = 79
    Call Siout(regaddr , Regdata)                           'enable CLK0, source=crystal, PLLA=integer mode, power level=8mA i.e. 0dB
  End If
  If Clk = 1 Then
    Siaddr1 = 34                                            'point to CLK1 base register addresses
    Siaddr2 = 50
    Siaddr3 = 52
    Sendfreq                                                'and send data to si5351 CLK1
    Regaddr = 17
    Regdata = 111
    Call Siout(regaddr , Regdata)                           'enable CLK1, source=crystal, PLLB=integer mode, power level=8mA i.e. 0dB
  End If
  If Clk = 2 Then
    Siaddr1 = 42                                            'point to CLK1 base register addresses
    Siaddr2 = 58
    Siaddr3 = 60
    Sendfreq                                                'and send data to si5351 CLK1
    Regaddr = 18
    Regdata = 111
    Call Siout(regaddr , Regdata)                           'enable CLK2, source=crystal, PLLB=integer mode, power level=8mA i.e. 0dB
  End If
End Sub

'**********************************************************************
' subroutine to initialise the si5351 chip via I2C
'**********************************************************************

Sub Si5351init

Regaddr = 183
Regdata = 192
Call Siout(regaddr , Regdata)                               'set xtal load cap to 10pF
Regaddr = 3
Regdata = 0
Call Siout(regaddr , Regdata)                               'enable all outputs
Regaddr = 15
Regdata = 0
Call Siout(regaddr , Regdata)                               'select xtal as PLL input source
Regaddr = 16
Regdata = 15                                                'PLLA is source, 8mA output
Call Siout(regaddr , Regdata)                               'configure clock 0 control register
Regaddr = 17
Regdata = 47                                                'PLLB is source, 8mA output
Call Siout(regaddr , Regdata)                               'configure clock 1 control register
Regaddr = 18
Regdata = 47                                                'PLLB is source, 8mA output
Call Siout(regaddr , Regdata)                               'configure clock 2 control register
Regaddr = 177
Waitms 10                                                   'wait 10ms just in case


End Sub

'********************************************************************************
' calculate data bytes for current VFO or BFO frequency to send to si5351a CLK0/1
'********************************************************************************
Sub Calculate

Rdivider = 1                                                'initialise r

Outdivider = 900000000 \ Gfreq                              'and outdivider

While Outdivider > 900
   Rdivider = 2 * Rdivider
   Outdivider = Outdivider \ 2
Wend

Fvco = Outdivider Mod 2                                     'check outdivider is an even number
If Fvco <> 0 Then Outdivider = Outdivider - 1               'using fvco as a temporary register

Fvco = Outdivider * Rdivider
Fvco = Fvco * Gfreq                                         'determine PLL frequency

Select Case Rdivider                                        'determine value of R44 bits [6:4]
      Case 1 : Rbyte = 0                                    '000 0000
      Case 2 : Rbyte = 16                                   '001 0000
      Case 4 : Rbyte = 32                                   '010 0000
      Case 8 : Rbyte = 48                                   '011 0000
      Case 16 : Rbyte = 64                                  '100 0000
      Case 32 : Rbyte = 80                                  '101 0000
      Case 64 : Rbyte = 96                                  '110 0000
      Case 128 : Rbyte = 112                                '111 0000
End Select

Afactor = Fvco \ Xtal
Fvalue = Afactor * Xtal                                     'fvalue = fvco -afactor * xtal
Fvalue = Fvco - Fvalue                                      'a+b/c
Fvalue = Fvalue * Cfactor
Fvalue = Fvalue \ Xtal
Bfactor = Fvalue

Ms0p1 = 128 * Outdivider
Ms0p1 = Ms0p1 - 512
Fvalue = 128 * Bfactor
Fvalue = Fvalue \ Cfactor
Finteger = Fvalue
Msnap1 = 128 * Afactor
Msnap1 = Msnap1 + Finteger                                  'finteger used instead of fvalue for variable matching
Msnap1 = Msnap1 - 512
Msnap3 = Finteger * Cfactor                                 'using msnap3 as temp result buffer
Msnap2 = 128 * Bfactor
Msnap2 = Msnap2 - Msnap3
Msnap3 = Cfactor

'calculations are complete so now build and send the required output data bytes by calling sendfreq  (About 20 bytes sent per frequency)

End Sub

'**********************************************************************
' send converted freq data to si5351 for CLK0 output
'**********************************************************************
Sub Sendfreq

Regaddr = Siaddr1
Call Siout(regaddr , Ms32)                                  'R26=msnap3[15:8]
Incr Regaddr
Call Siout(regaddr , Ms31)                                  'R27=msnap3[7:0]
Incr Regaddr
Call Siout(regaddr , Ms13)                                  'R28=msnap1[17:16] in R28[1:0] with R28[7:2]=0
Incr Regaddr
Call Siout(regaddr , Ms12)                                  'R29=msnap1[15:8]
Incr Regaddr
Call Siout(regaddr , Ms11)                                  'R30=msnap1[7:0]
Temp = Ms33                                                 'temp[3:0]=msnap3[19:16]
Swap Temp                                                   'temp[7:4]=msnap3[19:16]
Temp = Temp Or Ms23                                         'temp[7:4]=msnap3[19:16] and temp[3:0]=msnap2[19:16]
Incr Regaddr
Call Siout(regaddr , Temp)                                  'R31=(msnap3[19:16],msnap2[19:16])
Incr Regaddr
Call Siout(regaddr , Ms22)                                  'R32=msnap2[15:8]
Incr Regaddr
Call Siout(regaddr , Ms21)                                  'R33=msnap2[7:0]
Regaddr = Siaddr2
Regdata = 0
Call Siout(regaddr , Regdata)                               'R42=MS0_P3[15:8]
Incr Regaddr
Regdata = 1
Call Siout(regaddr , Regdata)                               'R43=MS0_P3[7:0]
Temp = Rbyte
Temp = Temp Or M0p13
Incr Regaddr
Call Siout(regaddr , Temp)                                  'R44=(0,R[7:4],MS0_P1[17:16])
Incr Regaddr
Call Siout(regaddr , M0p12)                                 'R45=MS0_P1[15:8]
Incr Regaddr
Call Siout(regaddr , M0p11)                                 'R46=MS0_P1[7:0]
Incr Regaddr
Regdata = 0
Call Siout(regaddr , Regdata)                               'R47=(MS0_P3[19:16],MS0_P2[19:16]) and both are always 0
Incr Regaddr
Regdata = 0
Call Siout(regaddr , Regdata)                               'R48=MS0_P2[15:8]=0 always
Incr Regaddr
Regdata = 0
Call Siout(regaddr , Regdata)                               'R49=MS0_P2[7:0]=0 always
If Outdivider = 4 Then                                      'for output frequencies greater than 150MHz and less than 200MHz
   Regaddr = Siaddr3
   Regdata = 12
   Call Siout(regaddr , Regdata)
   Incr Regaddr
   Regdata = 0
   Call Siout(regaddr , Regdata)
   Incr Regaddr
   Regdata = 0
   Call Siout(regaddr , Regdata)
End If
Regaddr = 3
Regdata = 0                                                 '4
Call Siout(regaddr , Regdata)                               'enable all outputs except CLK2

End Sub

'**********************************************************************
' i2c transmission to si5351
'**********************************************************************
Sub Siout(sireg As Byte , Sidata As Byte)
'subroutine sends databyte to register_nbr at i2c address 192 (si5351 register read is from address 194)
   'Write a single byte (slave address 192, register sireg, value sibyte)
    I2cstart
    I2cwbyte 192                                            'i2c address of si5351a
    Waitus 1                                                'suggested delay from forum
    I2cwbyte Sireg
    Waitus 1
    I2cwbyte Sidata
    Waitus 1
    I2cstop

End Sub


End